from threading import Event, Thread
import time
from typing import Optional

import serial
import simplepyble

from naneos.logger import LEVEL_INFO, get_naneos_logger
from naneos.partector_ble.partector_ble_device import PartectorBleDevice

logger = get_naneos_logger(__name__, LEVEL_INFO)


class SimplePyBleWrapper(Thread):
    SERVICE_UUID = "0bd51666-e7cb-469b-8e4d-2742f1ba77cc"
    CHAR_STD = "e7add780-b042-4876-aae1-112855353cc1"
    CHAR_AUX = "e7add781-b042-4876-aae1-112855353cc1"
    CHAR_WRITE = "e7add782-b042-4876-aae1-112855353cc1"
    CHAR_READ = "e7add783-b042-4876-aae1-112855353cc1"
    CHAR_SIZE_DIST = "e7add784-b042-4876-aae1-112855353cc1"

    def __init__(self):
        super().__init__()
        self._event = Event()
        self.initialized = False

        """Dict of all devices that are found in the scan method."""
        # [simplepyble.device, tuple[int, bytearray (advertisement)]]
        self._devices_to_check = {}
        """Dict containing all connected devices, that are in use."""
        self._partector_clients: dict[int, PartectorBleDevice] = {}

        self._adapter = self._init_get_first_adapter()
        if not self._adapter:
            return None

        self.initialized = True
        if self.initialized:
            self.start()

    def stop(self, blocking=True):
        self._event.set()
        if blocking:
            self.join()

    def run(self):
        while not self._event.is_set():
            self._loop_scan()
            self._loop_check_scan()
            time.sleep(1)

    def _init_get_first_adapter(self) -> Optional[simplepyble.Adapter]:
        adapters = simplepyble.Adapter.get_adapters()

        if not adapters:
            logger.error("No BLE adapter found")
            return None

        adapter = adapters[0]
        logger.info(f"Using BLE adapter: {adapter.identifier()}")

        return adapters[0]

    @staticmethod
    def get_full_advertisement(device) -> bytearray:
        adv = device.manufacturer_data()
        adv_0 = next(iter(adv.keys())).to_bytes(2, byteorder="little")
        adv_1 = next(iter(adv.values()))
        adv_bytes = bytearray(adv_0 + adv_1)

        return adv_bytes

    def _loop_scan(self) -> None:
        self._adapter.scan_for(850)

        devices = self._adapter.scan_get_results()

        for device in devices:
            if device.identifier() not in ["P2", "PartectorBT"]:
                continue
            #     logger.info(f"Device: {device.identifier()}")
            advertisement = self.get_full_advertisement(device)
            serial_number = PartectorBleDevice.check_naneos_adv(advertisement)

            if serial_number is None:
                continue

            self._devices_to_check[serial_number] = device

            logger.info(f"Data: {serial_number}")

    def _loop_check_scan(self) -> None:
        for sn, periphereal in self._devices_to_check.items():
            if sn in self._partector_clients:
                continue

            periphereal.connect()

            services = periphereal.services()
            service_characteristic_pair = []
            for service in services:
                for characteristic in service.characteristics():
                    service_characteristic_pair.append((service.uuid(), characteristic.uuid()))

                # for i, (service_uuid, characteristic) in enumerate(service_characteristic_pair):
                #     print(f"{i}: {service_uuid} {characteristic}")

            # logger.info(f"Connected to {sn}")
            contents = periphereal.notify(
                self.SERVICE_UUID, self.CHAR_STD, lambda data: print(data)
            )

            print(contents)


if __name__ == "__main__":
    ble = SimplePyBleWrapper()

    time.sleep(15)
    ble.stop()



from queue import Queue
from threading import Thread
from typing import Any, Callable, Optional

import serial

from naneos.logger.custom_logger import get_naneos_logger
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint
from naneos.serial_utils import list_serial_ports as ls_ports

logger = get_naneos_logger(__name__)


class ScanPartector(PartectorBluePrint):
    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 1,
    ) -> None:
        super().__init__(serial_number, port, verb_freq)

    def _init_serial_data_structure(self) -> None:
        """This field is not used in the scan partector, but mandatory in the partector blueprint."""
        pass

    def _serial_wrapper(self, func: Callable[[], Any]) -> Optional[Any]:
        """Wraps user func in try-except block. Forwards exceptions to the user."""
        if not self._connected:
            return None

        excep = "Was not able to fetch the serial number!"

        for _ in range(self.SERIAL_RETRIES):
            try:
                return func()
            except Exception as e:
                # ßlogger.error(f"SN{self._sn} Exception in _serial_wrapper: {e}")
                excep = f"SN{self._sn} Exception occured during user function call: {e}"

        raise Exception(excep)

    def _init_get_device_info(self) -> None:
        try:
            if self._sn is None:
                self._sn = self._get_serial_number_secure()
            self._fw = self.get_firmware_version()
            logger.debug(f"Connected to SN{self._sn} on {self._port}")
        except Exception:
            # this gets just passed because scan partector scans every port
            # therefore it does not know if there is a partector
            pass

    def _set_verbose_freq(self, freq: int = 0) -> None:
        """This field is only used to set the verbose frequency to 0"""
        self._write_line("X0000!")

    def _check_connection(self) -> None:
        """Checks if there is a serial connection."""
        if isinstance(self._ser, serial.Serial) and self._ser.is_open:
            self._connected = True


def scan_for_serial_partector(
    serial_number: int, partector_version: str
) -> Optional[str]:
    """Scans all possible ports using threads (fast)."""
    threads = []
    q_1 = Queue()
    q_2 = Queue()
    q_2_pro = Queue()
    q_2_pro_cs = Queue()

    [
        threads.append(
            Thread(target=__scan_port, args=(port, q_1, q_2, q_2_pro, q_2_pro_cs))
        )
        for port in ls_ports()
    ]
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    q = Queue()
    if partector_version == "P1":
        q = q_1
    elif partector_version == "P2":
        q = q_2
    elif partector_version == "P2pro":
        q = q_2_pro
    elif partector_version == "P2proCS":
        q = q_2_pro_cs

    ports = {k: v for d in tuple(q.queue) for (k, v) in d.items()}
    logger.debug(f"Found ports: {ports}")

    if serial_number in ports:
        return ports[serial_number]

    return None


def scan_for_serial_partectors(ports_exclude: Optional[list] = None) -> dict:
    """Scans all possible ports using threads (fast)."""
    threads = []
    q_1, q_2, q_2_pro, q_2_pro_cs = (Queue() for _ in range(4))

    if ports_exclude is None:
        ports_exclude = []

    [
        threads.append(
            Thread(target=__scan_port, args=(port, q_1, q_2, q_2_pro, q_2_pro_cs))
        )
        for port in ls_ports(ports_exclude)
        if port not in ports_exclude
    ]
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    p1 = {k: v for x in tuple(q_1.queue) for (k, v) in x.items()}
    p2 = {k: v for x in tuple(q_2.queue) for (k, v) in x.items()}
    p2_pro = {k: v for x in tuple(q_2_pro.queue) for (k, v) in x.items()}
    p2_pro_cs = {k: v for x in tuple(q_2_pro_cs.queue) for (k, v) in x.items()}

    return {"P1": p1, "P2": p2, "P2pro": p2_pro, "P2proCS": p2_pro_cs}


def __scan_port(
    port: str, q_1: Queue, q_2: Queue, q_2_pro: Queue, q_2_pro_cs: Queue
) -> None:
    try:
        partector = ScanPartector(port=port)

        if partector._sn is None:
            pass
        elif partector._sn < 1000:
            q_1.put({partector._sn: port})
        else:
            if partector._fw < 310:
                q_2.put({partector._sn: port})
            else:
                name: str = partector.write_line("name?")[1]
                if name == "P2":
                    q_2.put({partector._sn: port})
                elif name == "P2pro":
                    q_2_pro.put({partector._sn: port})
                elif name == "P2proCS":
                    q_2_pro_cs.put({partector._sn: port})

        partector.close(blocking=True)
    except Exception:
        if isinstance(partector, ScanPartector):
            partector.close(blocking=True)


if __name__ == "__main__":
    print(scan_for_serial_partectors())



from naneos.serial_utils import list_serial_ports


def test_check_serial_connection() -> None:
    """Test if the serial connection is working."""
    ports: list[str] = list_serial_ports()
    assert len(ports) > 0, "It seems you have no device connected to your computer."



import numpy as np
import pandas as pd

from naneos.iotweb import Partector2ProGarageUpload


def _callback_upload(state: bool) -> None:
    print(f"Upload state: {state}")


def print_ser(ser: pd.Series) -> None:
    print(type(ser.name))


def test_upload_cs() -> None:
    df = pd.read_pickle("tests/df_upload_test.pkl").iloc[0:1]
    # set first index to nan
    # df.index = np.nan
    # print(df)
    # print(df.describe())

    df_append = pd.DataFrame(columns=df.columns, index=[np.NaN])
    df = pd.concat([df, df_append])

    # print(df.dtypes)

    # make index int
    df.dropna(inplace=True)
    df.index = df.index.astype(int)

    df.apply(print_ser, axis=1)

    # Create a Partector2ProGarageUpload object
    # Partector2ProGarageUpload(df, 8448, _callback_upload).start()


if __name__ == "__main__":
    test_upload_cs()



from naneos.iotweb.download.downloader import download_from_iotweb
from naneos.iotweb.iotweb import P1uploadThread
from naneos.iotweb.partector2_pro_garage_upload import Partector2ProGarageUpload

__all__ = ["P1uploadThread", "Partector2ProGarageUpload", "download_from_iotweb"]



from naneos.iotweb import download_from_iotweb


def test_download_8134() -> None:
    import datetime as dt
    import os

    import pandas as pd

    token: str | None = os.getenv("IOT_GUEST_TOKEN", None)
    if token is None:
        raise ValueError("No token found in your environment")

    # use local timezone for start
    start = dt.datetime(2023, 11, 26)
    stop = dt.datetime(2023, 11, 27)
    serial_number = "8134"
    name = "iot_guest"

    df: pd.DataFrame = download_from_iotweb(name, serial_number, start, stop, token)
    assert df.shape[0] > 0



import time

from naneos.partector import Partector2Pro, scan_for_serial_partectors


def test_partector2_pro() -> None:
    """Test if the serial connection is working 10 times."""
    partectors = scan_for_serial_partectors()
    p2_pro = partectors["P2pro"]

    assert p2_pro, "No Partector found!"

    serial_number = next(iter(p2_pro.keys()))

    for _ in range(33):
        p2 = Partector2Pro(serial_number=serial_number)
        p2.close(verbose_reset=False)



from naneos.partector.partector1 import Partector1
from naneos.partector.partector2 import Partector2
from naneos.partector.partector2_pro import Partector2Pro
from naneos.partector.partector2_pro_cs import Partector2ProCs
from naneos.partector.scan_for_partector import (
    scan_for_serial_partector,
    scan_for_serial_partectors,
)

__all__ = [
    "Partector1",
    "Partector2",
    "Partector2Pro",
    "Partector2ProCs",
    "scan_for_serial_partector",
    "scan_for_serial_partectors",
]



from typing import Optional

from naneos.partector.blueprints._data_structure import (
    PARTECTOR2_DATA_STRUCTURE_V320,
    PARTECTOR2_PRO_DATA_STRUCTURE_V311,
    PARTECTOR2_PRO_DATA_STRUCTURE_V336,
)
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint


class Partector2Pro(PartectorBluePrint):
    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 6,
        hw_version: str = "P2pro",
    ) -> None:
        super().__init__(serial_number, port, verb_freq, hw_version)

    def _init_serial_data_structure(self) -> None:
        """This gets passed here and is set in the set_verbose_freq method."""
        # TODO: My idea is to remove that completely and set the data structure in the set_verbose_freq method for all devices
        pass

    def _set_verbose_freq(self, freq: int) -> None:
        if freq == 0:
            self._write_line("X0000!")
        elif freq in [1, 2, 3]:  # std p2 mode
            if self._fw >= 311:
                self._data_structure = PARTECTOR2_DATA_STRUCTURE_V320

            self._write_line("h2001!")  # activates harmonics output
            self._write_line("M0000!")  # deactivates size dist mode
            self._write_line(f"X000{freq}!")
        elif freq == 6:  # p2 pro mode
            if self._fw >= 336:
                self._data_structure = PARTECTOR2_PRO_DATA_STRUCTURE_V336
            else:
                self._data_structure = PARTECTOR2_PRO_DATA_STRUCTURE_V311

            self._write_line("X0006!")  # activates verbose mode
            self._write_line("h2001!")  # activates harmonics output
            self._write_line("M0004!")  # activates size dist mode
            self._write_line("A0001!")  # activates the antispikes


if __name__ == "__main__":
    import time

    from naneos.partector import scan_for_serial_partectors

    partectors = scan_for_serial_partectors()
    p2_pro = partectors["P2pro"]

    assert p2_pro, "No Partector found!"

    serial_number = next(iter(p2_pro.keys()))
    p2 = Partector2Pro(serial_number=serial_number)

    for _ in range(3):
        time.sleep(10)
        df = p2.get_data_pandas()
        print(df[["particle_number_10nm", "particle_number_16nm"]])

    p2.close(verbose_reset=False, blocking=True)



import asyncio
from typing import Optional

from bleak import BleakClient, BleakScanner
from bleak.backends.device import BLEDevice


async def discover() -> Optional[BLEDevice]:
    device: BLEDevice | None = await BleakScanner.find_device_by_name("P2", timeout=4.0)  # type: ignore

    if isinstance(device, BLEDevice):
        return device

    return None


async def print_client_services(device: BLEDevice) -> None:
    try:
        async with BleakClient(device, timeout=1) as client:
            for service in client.services:
                print(f"[Service] {service}")
                for characteristic in service.characteristics:
                    print(f"[Characteristic] {characteristic}")
    except Exception as e:
        print("Error:", e)


async def find_and_print_services() -> None:
    device: Optional[BLEDevice] = await discover()

    if isinstance(device, BLEDevice):
        print(f"Found device: {device}")
        await print_client_services(device)
    else:
        print("No device found")


if __name__ == "__main__":
    asyncio.run(find_and_print_services())



import time

from naneos.partector import Partector1, Partector2, scan_for_serial_partectors


def test_readme_example():
    # Lists all available Partector2 devices
    x = scan_for_serial_partectors()

    print(
        x
    )  # eg. {'P1': {}, 'P2': {8112: '/dev/cu.usbmodemDOSEMet_1'}, 'P2pro': {}, 'P2proCS': {}}

    # Split dictionary into P1 and P2 devices
    p1 = x["P1"]
    p2 = x["P2"]
    p2_pro = x["P2pro"]

    if len(p1) > 0:
        print("Found Partector1 devices:")
        for k, v in p1.items():
            print(f"Serial number: {k}, Port: {v}")

        # Connect to the first device with sn
        p1_dev = Partector1(serial_number=next(iter(p1.keys())))
        # or with port
        # p1_dev = Partector1(port=next(iter(p1.values())))

        time.sleep(2)

        # Get the data as a pandas DataFrame
        data = p1_dev.get_data_pandas()
        print(data)

        p1_dev.close()

    if len(p2) > 0:
        print("Found Partector2 devices:")
        for k, v in p2.items():
            print(f"Serial number: {k}, Port: {v}")

        # Connect to the first device with sn
        p2_dev = Partector2(serial_number=next(iter(p2.keys())))
        # or with port
        # p2_dev = Partector2(port=next(iter(p2.values())))

        time.sleep(2)

        # Get the data as a pandas DataFrame
        data = p2_dev.get_data_pandas()
        print(data)

        p2_dev.close()


if __name__ == "__main__":
    test_readme_example()
